


\section{Generic testing}

The idea is to have test cases that work for all of the
implementations of ESF arrays.  Therefore the test data needs to be
generic.  There may be more specific test data for specific
implementations.

\subsection{Defining operations for testing}

print a segment of an array a from index i to index j, using machine
state mst and an implementation of the lookup function.

\begin{code}
printSegment flookup mst a i j =
  printArrayValue flookup (i,j) mst a
\end{code}


To test, keep a list of all the extant arrays.

The Operation type is for testing.  An array is referenced by a
number (an Avarnum), which is an index into a list of extant arrays.
Naturally this is not a useful way to keep hold of all the arrays
in a real program, but it provides a handy way to test updates and
lookups.  The numbering works as follows.  There is a list of
arrays, and an array is specified by an Int which is an index into
this list.  When an array is updated, the newly created array
number is attached to the end of the list.  When an array is
deallocated, its number is removed from the list.

\begin{code}
newtype Avarnum = Avarnum Int   deriving (Eq,Read,Show)

data Operation
  = Update Avarnum EltIndex EltValue
  | Lookup Avarnum EltIndex
  | MemDump
  deriving (Eq,Read,Show)

showOp (Update (Avarnum an) i v) =
  "Update a" ++ show an ++ " [" ++ show i ++ "] := " ++ show v
showOp (Lookup (Avarnum an) i) =
  "Lookup a" ++ show an ++ " [" ++ show i ++ "]"
showOp MemDump = "MemDump"

data OperationList
  = EmptyOpList
  | HaveOperation Operation OperationList
  deriving (Eq,Read,Show)

showOpList ops =
  concat (map ((++"\n") . showOp) (opListToList ops))

showOpsData (OpsData ops) = showOpList ops

data OpsData = OpsData OperationList
  deriving (Eq,Read,Show)

opsDataSize :: OpsData -> Int
opsDataSize (OpsData ops) = length (opListToList ops)


listToOpList :: [Operation] -> OperationList
listToOpList [] = EmptyOpList
listToOpList (x:xs) = HaveOperation x (listToOpList xs)

opListToList :: OperationList -> [Operation]
opListToList EmptyOpList = []
opListToList (HaveOperation x xs) = x : opListToList xs

data OperationResult
  = UpdateResult Avarnum EltIndex EltValue Avarnum
  | LookupResult Avarnum EltIndex (Maybe EltValue)
  | MemDumpResult [[(EltIndex, Maybe EltValue)]]
  deriving (Eq,Read,Show)

opavarnum :: Operation -> Int
opavarnum (Update (Avarnum n) _ _) = n
opavarnum (Lookup (Avarnum n) _) = n

showOpRes :: OperationResult -> String
showOpRes (UpdateResult an ai av rn) =
  "Update: Array" ++ show rn ++ " = update Array"
    ++ show an ++ " [" ++ show ai ++ "] := " ++ show av
showOpRes (LookupResult an ai mv) =
  "Lookup: Array" ++ show an ++ " [" ++ show ai ++ "] = " ++ show mv
showOpRes (MemDumpResult xs) =
  "MemDump: ********" ++ show xs

showOperationResults :: [OperationResult] -> String
showOperationResults xs = concat (map f xs)
  where f x = showOpRes x ++ "\n"

test_data_1 :: [Operation]
test_data_1 =
  [Update (Avarnum 0) (50) (500),
   Update (Avarnum 0) (51) (600),
   Update (Avarnum 1) (52) (700),
   Lookup (Avarnum 1) (50),
   Lookup (Avarnum 2) (51),
   Lookup (Avarnum 3) (50),
   Lookup (Avarnum 3) (51),
   Lookup (Avarnum 3) (52)]
 \end{code}
 
The initial alist contains (0,0) indicating that there is already
an array empty, with index number 0.  The empty array is always
present, and it's important to ensure that the deallocate operation
can never remove it.

\subsection{Running test case monadically}


These may be redundant by now...

type MachineState = TreeMachine PtcCellState
type SpecState = [(Int,Int)]
type RunState = (SpecState,MachineState)




\subsection{Running test case in pure non-monadic style}

NOT IN PROGRAM:

\begin{verbatim}
sensible_Avarnums :: [Operation] -> Bool
sensible_Avarnums xs = f 1 xs
  where f n [] = True
        f n (x:xs) =
           let av = opavarnum x
               ok = 0 <= av && av < n
           in ok && f (n+1) xs
\end{verbatim}

Run a test script using an implementation of esfa

\begin{code}
run fupdate flookup testscript =
  runstepsPure fupdate flookup 1 [(0,0)] [] testscript
\end{code}

\begin{code}
test_runsteps =
  do putStrLn (take 50 (repeat '-'))
     putStrLn "Running test_data_1"

     putStrLn "Pure version with DP1, test_data_1"
     putStrLn (showOperationResults
                (runstepsPure dp1update dp1lookup
                   1 [(0,0)] [] test_data_1))
     putStrLn (take 50 (repeat '-'))

     putStrLn "Pure version with Spec, test_data_1"
     putStrLn (showOperationResults
                (runstepsPure listUpdatewithSt listLookupwithSt
                   1 [(0,[])] [] test_data_1))
     putStrLn (take 50 (repeat '-'))

     putStrLn "Compare spec/DP1 with test_data_1"
     putStrLn (show (prop_meetSpecDP1 (OpsData (listToOpList test_data_1))))
\end{code}

\begin{code}
runstepsPure
  :: (s -> a -> (EltIndex,EltValue) -> (s,a))   -- update fcn
  -> (s -> a -> EltIndex -> Maybe EltValue)     -- lookup fcn
  -> Int                                        -- nextAnum
  -> [(Int,a)]                                  -- alist
  -> s                                          -- state
  -> [Operation]                                -- operations to run
  -> [OperationResult]                        -- results

runstepsPure fupdate flookup nextAnum alist st [] = []

runstepsPure fupdate flookup nextAnum alist st (Update (Avarnum j0) i v : xs) =
  let j = (abs j0) `mod` (length alist)
      (aj,an) = alist !! j
      (st',a') = fupdate st an (i,v)
      alist' = alist ++ [(nextAnum,a')]
      opres1 = UpdateResult (Avarnum j) i v (Avarnum nextAnum)
      y = memDump flookup
            (50, 54) -- ????????????
            st' (map snd alist') -- ???????put st' adding ' caused exception
      opres2 = MemDumpResult y
  in  opres1 : opres2 :
        runstepsPure fupdate flookup (nextAnum+1) alist' st' xs

runstepsPure fupdate flookup nextAnum alist st (Lookup (Avarnum j0) i : xs) =
  let j = (abs j0) `mod` (length alist)
      (aj,an) = alist !! j
      maybeV = flookup st an i
      opres = LookupResult (Avarnum j) i maybeV
  in opres : runstepsPure fupdate flookup nextAnum alist st xs

runstepsPure fupdate flookup nextAnum alist st (MemDump : xs) =
  let y = memDump flookup
            (50, 54) -- ????????????
            st (map snd alist)
      opres = MemDumpResult y
  in opres : runstepsPure fupdate flookup nextAnum alist st xs

type Idx = Int
\end{code}

\begin{code}
showDp1update (s,a) lbl =
  do putStrLn ("update " ++ lbl ++ " => " ++ show a)
     putStrLn ("state = " ++ show s)
     dumpStateDP1 s lbl
     putStrLn ""
\end{code}


\subsection{Manual testing}
\label{sec:manual-test-cases}



\section{Automated testing with Quickcheck}

\subsection{Correctness properties}

\begin{code}
prop_meetSpecDP1 :: OpsData -> Bool
prop_meetSpecDP1 (OpsData ops) =
  let xs = opListToList ops
  in ((runstepsPure dp1update dp1lookup 1 [(0,0)] [] xs
          ==
        runstepsPure listUpdatewithSt listLookupwithSt 1 [(0,[])] [] xs))

cprop_meetSpecDP1 :: OpsData -> Property
cprop_meetSpecDP1 x =
  collect (opsDataSize x) (prop_meetSpecDP1 x)

\end{code}

\subsection{Generating operations for QuickCheck}

\begin{code}

instance Arbitrary Avarnum where
  arbitrary = sized arbAvar

arbAvar n =
  do x <- choose (0,n-1)
     return (Avarnum x)

instance Arbitrary Operation where
  arbitrary = sized arbOp

data AlistSizeChange
  = IncAlistSize
  | NoChangeAlistSize
  | DecAlistSize
  deriving (Eq,Read,Show)

arbOp n =
  oneof [do av <- arbAvar n
            idx <- arbitrary
            val <- arbitrary
            return (Update av idx val),
         do av <- arbAvar n
            idx <- arbitrary
            return (Lookup av idx)]

fooarbOp n =
  oneof [do av <- arbAvar n
            idx <- arbitrary
            val <- arbitrary
            return (IncAlistSize, Update av idx val),
         do av <- arbAvar n
            idx <- arbitrary
            return (NoChangeAlistSize, Lookup av idx)]

mkOperationList n alistsize
  | n<=0 = return EmptyOpList
  | n>0 =
      do (change,op) <- fooarbOp alistsize
         let alistsize' = case change of
                            IncAlistSize -> alistsize + 1
                            NoChangeAlistSize -> alistsize
                            DecAlistSize -> alistsize -1
         ops <- mkOperationList (n-1) alistsize'
         return (HaveOperation op ops)
      

instance Arbitrary OpsData where
  arbitrary = sized arbOpsData

arbOpsData n =
  do r <- mkOperationList n 1
     return (OpsData r)



testQC =
  do putStrLn "testQC for DP1"
     putStrLn "EltIndex"
     sample (arbitrary :: Gen EltIndex)
     putStrLn "EltIndex"
     sample (arbitrary :: Gen EltValue)
     putStrLn "Avarnum"
     sample (arbitrary :: Gen Avarnum)
     putStrLn "Operation"
     sample (arbitrary :: Gen Operation)
     putStrLn "Operations"
     sample (arbitrary :: Gen [Operation])

     putStrLn (take 50 (repeat '-'))
     putStrLn "quickCheck prop_meetSpecDP1"
     quickCheck prop_meetSpecDP1
     putStrLn (take 50 (repeat '-'))
\end{code}


\subsection{Main program}
   
Here is the main program.

 took out     testlist

\begin{code}
maintest :: IO ()
maintest =
  do putStrLn "hello, world"
     test_ax

     mytest1
     test_runsteps

     putStrLn (take 50 (repeat '-'))
     putStrLn "quickCheck prop_meetSpecDP1"
     let qc_args = stdArgs {maxSize = 200, chatty = True}

     quickCheckWith qc_args cprop_meetSpecDP1
     putStrLn (take 50 (repeat '-'))
\end{code}

\begin{code}
tryTests :: IO ()
tryTests =
  do putStrLn "tryTests"
     xs <- sample' (arbitrary :: Gen OpsData)
     doall xs

doall [] = return ()
doall (x:xs) =
  do putStrLn (take 50 (repeat '-'))
     putStrLn (show x)
     putOpsData x
     putStrLn ("number of operations = " ++ show (opsDataSize x))
     doall xs

putOpsData :: OpsData -> IO ()
putOpsData x =
  do putStrLn "OpsData..."
     putStrLn (showOpsData x)
\end{code}

\section{Conclusion}

Here are the results and observations.

\bibliography{srcEsfa}

\end{document}

------------------------------------------------------------------------
From here on, there is old code not currently being used
------------------------------------------------------------------------

Code from subsection{Running test case monadically}
{-
runmain :: Int -> [Operation] -> IO ()
runmain = runtreespec runlooper
-}

{-
runtreespec :: StateT RunState IO () -> Int -> [Operation] ->  IO ()
runtreespec f k ops =
  evalStateT (f ops) (initSpec, mkTree initPtcCell k)
-}

initSpec = []

{-
runlooper :: [Operation] -> StateT RunState IO ()
runlooper [] = return ()
runlooper (x:xs) =
  do case x of
       Update anum i v ->
         do a' <- ptcUpdate anum (i,v)
            liftIO $ putStrLn ("Update returnes " ++ show a')
       Lookup anum i ->
         do v <- ptcLookup anum i
            liftIO $ putStrLn ("Lookup returnes " ++ show v)
       _ ->   liftIO $ putStrLn "Unimplemented operation"
     runlooper xs
-}
\end{code}

\section{Cruft}

This stuff will probably be deleted.

\subsection{Tree Specification with explicit state (no monad)}

Use tree machine without monad - this is too painful and seems like a bad
idea.  Monad is better.

\begin{verbatim}

-- the tree machine state
type TMST = TreeMachine PtcCellState


ptce_initNames :: TMST -> TMST
ptce_initNames tmst = fst (tscanl f g (+) 0 tmst)
  where f c = 1
        g c i = c {ptcCellId = AName i,
                   ptcMapDef = i==0}

ptce_showState :: TMST -> String
ptce_showState tmst =
  concat [show i ++ ". " ++ showPtcCell x ++ "\n"
           | (i,x) <- zip [0..] (getTree tmst)]

ptce_encode :: TMST -> AName -> Maybe ACode
ptce_encode tmst (AName n)
  | n==0 = Just (ACode 0)
  | n/=0 = tfold1 f g tmst
  where f x = if AName n == ptcCellId x then Just (ptcMapCode x) else Nothing
        g Nothing Nothing = Nothing
        g (Just c) Nothing = Just c
        g Nothing (Just c) = Just c
        g (Just c) (Just d) = Just c  -- impossible, would indicate error

-- message = (name found to left, value of new name)
ptce_NewName :: TMST -> ACode -> (TMST, (Bool,AName))
ptce_NewName tmst newcode = tscanl f g h (False, AName 0) tmst
  where f c = (not (ptcMapDef c), ptcCellId c)
        g c (l,n) =
          let foundThis = not (ptcMapDef c) && not l
              newMapDef = ptcMapDef c || foundThis
              newMapCode = if foundThis then newcode else ptcMapCode c
          in c {ptcMapDef = newMapDef, ptcMapCode = newMapCode}
        h (l1,n1) (l2,n2) = (l1 || l2, if l1 then n1 else n2)  

-- message means empty cell found
ptce_NewCell :: TMST -> (TMST, Bool)
ptce_NewCell tmst = tscanl f g (||) False tmst
  where f c = not (ptcEltDef c)
        g c True = c {ptcSelect = False}
        g c False =
          if ptcEltDef c
            then c  {ptcSelect = False}
            else c {ptcSelect = True}


ptce_Update
  :: TMST
  -> AName
  -> (EltIndex,EltValue)
  -> (TMST, Maybe AName)
ptce_Update tmst1 aname (aidx,aval) =
 let maybeacode = ptce_encode tmst1 aname
 in case maybeacode of
       Nothing -> Nothing --  "update: no code for array"
       Just (ACode acode) ->
         let
-- Calculate code of new array
             acode' = acode + 1
-- Calculate name of new array and store its code in name/code map
            (namefound, resultname) = ptce_NewName tmst1 (ACode acode')
-- Allocate cell for new element
            (tmst2,cellfound) = ptce_NewCell tmst1
         in case cellfound of
              False -> return Nothing -- "failed to allocate new cell"
              True ->
                let tmst3 =   -- Adjust existing elements
                      mtmap (\c ->
                              c { ptcLow =
                                    if ptcEltDef c && ptcLow c > acode
                                      then ptcLow c + 1
                                      else ptcLow c,
                                  ptcHigh =
                                    if ptcHigh c >= acode
                                      then ptcHigh c + 1)
                                      else ptcHigh c,
                                  ptcMapCode =
                                      if (ptcCellId c /= resultname)
                                               -- already set code in result
                                           && (ptcMapCode c > acode)
                                        then ptcMapCode c + 1)
                                        else ptcMapCode c })

-- Store contents of new element in the cell that was allocated
                    low = ACode acode'
                    high = ACode acode'
                    tmst4 =
                      mtmap (\c ->
                              if ptcSelect c
                                then c {ptcEltDef = True,
                                        ptcLow = low,
                                        ptcHigh = high,
                                        ptcInd = aidx,
                                        ptcValField = aval}
                                else c)
                in (Just resultname)
\end{verbatim}

\subsection{Deleted code}

\begin{verbatim}
instance Arbitrary OperationList where
  arbitrary =
    do (d1,op1) <- fooarbOp 1
       (d2,op2) <- fooarbOp 10
       return (HaveOperation
                 op1
                 (HaveOperation op2 EmptyOpList))
\end{verbatim}

\subsection{Temporary experiments}

What follows are just some experiments to see how to do things,
this doesn't belong in the paper at all, not even in appendices.

\subsubsection{QuickCheck}

\begin{code}

myreverse :: [a] -> [a]
myreverse [] = []
myreverse (x:xs) = myreverse xs ++ [x]

prop_RevApp :: [Int] -> [Int] -> Bool
prop_RevApp xs ys =
  myreverse (xs++ys) == myreverse ys ++ myreverse xs

\end{code}

\subsubsection{Ad hoc testing}

A quick tool for unit testing.

\begin{code}
check :: (Eq a, Show a) => a -> a -> IO ()
check expect actual =
  do putStr (if expect == actual then "ok " else "ERROR ")
     putStrLn (show actual)
     return ()
\end{code}


\subsubsection{Typesetting experiments}

Experiment with typesetting in lhs2TeX.

\begin{code}
foobar :: Int
foobar = 42
\end{code}

{-  Change: using the cellid as the array name
-- message = (name found to left, value of new name)
ptcNewName :: ACode -> StateT (PtcRunState a) IO (Bool,AName)
ptcNewName newcode =
  mtscanl f_NewName g_NewName h_NewName (False, 0)
  where
    f_NewName c = (not (ptcMapDef c), ptcCellId c)
    g_NewName c (l,n) =
      let foundThis = not (ptcMapDef c) && not l
          newMapDef = ptcMapDef c || foundThis
          newMapCode = if foundThis then newcode else ptcMapCode c
      in c {ptcMapDef = newMapDef, ptcMapCode = newMapCode}
    h_NewName (l1,n1) (l2,n2) = (l1 || l2, if l1 then n1 else n2)  
-}
